# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.enums import SleepTimingType, TakeScreenshotConditionType


class StepSettings:
    """Represents settings for automatic step reporting."""
    def __init__(self, sleep_time: int = 0, sleep_timing_type: SleepTimingType = None, timeout: int = -1,
                 invert_result: bool = False,
                 screenshot_condition: TakeScreenshotConditionType = TakeScreenshotConditionType.Failure):
        self.sleep_time = sleep_time
        self.sleep_timing_type = sleep_timing_type
        self.timeout = timeout
        self.invert_result = invert_result
        self.screenshot_condition = screenshot_condition



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket
import atexit

from urllib.parse import urlparse
from src.testproject.sdk.exceptions import AgentConnectException


class SocketManager:
    """Class used to manage the development TCP socket connection.

    Attributes:
        __instance (SocketManager): The singleton instance of this class
        __socket (socket.Socket): A singleton socket instance used to communicate with the Agent
    """

    __instance = None

    __socket = None

    def __init__(self):
        """Create SocketManager instance and register shutdown hook"""
        atexit.register(self.close_socket)

    @classmethod
    def instance(cls):
        """Return the singleton instance of the SocketManager class"""
        if cls.__instance is None:
            cls.__instance = SocketManager()
        return cls.__instance

    def close_socket(self):
        """Close the connection to the Agent development socket"""
        if self.is_connected():
            try:
                SocketManager.__socket.shutdown(socket.SHUT_RDWR)
                SocketManager.__socket.close()
                SocketManager.__socket = None
                logging.info("Connection to Agent closed successfully")
            except socket.error as msg:
                logging.error(f"Failed to close socket connection to Agent: {msg}")

    def open_socket(self, socket_address: str, socket_port: int):
        """Opens a connection to the Agent development socket

        Args:
            socket_address (str): The address for the socket
            socket_port (int): The development socket port to connect to
        """

        if SocketManager.__socket is not None:
            logging.debug("open_socket(): Socket already exists")
            return

        if self.is_connected():
            logging.debug("open_socket(): Socket is already connected")
            return

        host = urlparse(socket_address).hostname

        SocketManager.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        SocketManager.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        SocketManager.__socket.connect((host, socket_port))

        if not self.is_connected():
            raise AgentConnectException("Failed connecting to Agent socket")

        logging.info(
            f"Socket connection to {host}:{socket_port} established successfully"
        )

    def is_connected(self) -> bool:
        """Sends a simple message to the socket to see if it's connected

            Returns:
                bool: True if the socket is connected, False otherwise
        """
        if SocketManager.__socket is None:
            return False

        try:
            SocketManager.__socket.send("test".encode("utf-8"))
            return True
        except socket.error as msg:
            logging.warning(f"Socket not connected: {msg}")
            return False



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from appium.webdriver.webdriver import WebDriver as AppiumWebDriver

from src.testproject.classes import StepSettings
from src.testproject.enums import EnvironmentVariable
from src.testproject.helpers import ReportHelper, LoggingHelper, ConfigHelper, AddonHelper
from src.testproject.rest import ReportSettings
from src.testproject.sdk.exceptions import SdkException
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers import CustomAppiumCommandExecutor
from src.testproject.sdk.internal.reporter import Reporter
from src.testproject.sdk.internal.session import AgentSession


class Remote(AppiumWebDriver):
    """Used to create a new Android driver instance

    Args:
        desired_capabilities (dict): Automation session desired capabilities and options
        token (str): Developer token to be used to communicate with the Agent
        projectname (str): Project name to report
        jobname (str): Job name to report
        disable_reports (bool): set to True to disable all reporting (no report will be created on TestProject)

    Attributes:
        _desired_capabilities (dict): Automation session desired capabilities and options
        _agent_client (AgentClient): client responsible for communicating with the TestProject agent
        _agent_session (AgentSession): stores properties of the current agent session
        command_executor (CustomAppiumCommandExecutor): the HTTP command executor used to send commands
        w3c (bool): indicates whether or not the driver instance uses the W3C dialect
        session_id (str): contains the current session ID
    """

    __instance = None

    def __init__(
        self,
        desired_capabilities: dict = None,
        token: str = None,
        projectname: str = None,
        jobname: str = None,
        disable_reports: bool = False,
    ):
        if Remote.__instance is not None:
            raise SdkException("A driver session already exists")

        LoggingHelper.configure_logging()

        self._desired_capabilities = desired_capabilities

        self._token = token if token is not None else ConfigHelper.get_developer_token()

        if disable_reports:
            # Setting the project and job name to empty strings will cause the Agent to not initialize a report
            self._projectname = ""
            self._jobname = ""
        else:
            self._projectname = (
                projectname
                if projectname is not None
                else ReportHelper.infer_project_name()
            )
            self._jobname = (
                jobname if jobname is not None else ReportHelper.infer_job_name()
            )

        reportsettings = ReportSettings(self._projectname, self._jobname)

        self._agent_client: AgentClient = AgentClient(
            token=self._token,
            capabilities=self._desired_capabilities,
            reportsettings=reportsettings,
        )
        self._agent_session: AgentSession = self._agent_client.agent_session
        self.w3c = True if self._agent_session.dialect == "W3C" else False

        AppiumWebDriver.__init__(
            self,
            command_executor=self._agent_session.remote_address,
            desired_capabilities=self._desired_capabilities,
        )

        self.command_executor = CustomAppiumCommandExecutor(
            agent_client=self._agent_client,
            remote_server_addr=self._agent_session.remote_address,
        )

        # this ensures that mobile-specific commands are also available for our command executor
        self._addCommands()

        Remote.__instance = self

    @classmethod
    def instance(cls):
        """Returns the singleton instance of the driver object"""
        return Remote.__instance

    @property
    def step_settings(self):
        return self.command_executor.settings

    @step_settings.setter
    def step_settings(self, step_settings: StepSettings):
        self.command_executor.settings = step_settings

    def start_session(self, capabilities, browser_profile=None):
        """Sets capabilities and sessionId obtained from the Agent when creating the original session."""
        self.session_id = self._agent_session.session_id
        logging.info(f"Session ID is {self.session_id}")

    def report(self) -> Reporter:
        """Enables access to the TestProject reporting actions from the driver object"""
        return Reporter(self.command_executor)

    def addons(self) -> AddonHelper:
        """Enables access to the TestProject addon execution actions from the driver object

        Returns:
            AddonHelper: object giving access to addon proxy methods
        """
        return AddonHelper(self._agent_client)

    def quit(self):
        """Quits the driver and stops the session with the Agent, cleaning up after itself."""
        # Report any left over driver command reports
        self.command_executor.clear_stash()

        # Make instance available again
        Remote.__instance = None

        try:
            AppiumWebDriver.quit(self)
        except Exception:
            pass

        # Stop the Agent client
        self.command_executor.agent_client.stop()

        # Clean up any environment variables set in the decorator
        for env_var in [
            EnvironmentVariable.TP_TEST_NAME,
            EnvironmentVariable.TP_PROJECT_NAME,
            EnvironmentVariable.TP_JOB_NAME,
        ]:
            EnvironmentVariable.remove(env_var)



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union

from src.testproject.classes import StepSettings
from src.testproject.enums import SleepTimingType, TakeScreenshotConditionType
from src.testproject.sdk.drivers.webdriver import Remote
from src.testproject.sdk.drivers.webdriver.base import BaseDriver


class DriverStepSettings:
    """Implementation of the 'with' compound statement.

    If driver auto step reports is enabled, all the driver commands within this compound statement will be reported
    according to the given StepSettings.
    """
    def __init__(self, driver: Union[BaseDriver, Remote], step_settings: StepSettings):
        """Initializes the 'with' statement."""
        self.previous_settings = driver.command_executor.settings
        self.driver = driver
        # If inherit take the previous step settings.
        if step_settings.sleep_timing_type and step_settings.sleep_timing_type is SleepTimingType.Inherit:
            step_settings.sleep_timing_type = self.previous_settings.sleep_timing_type
        if step_settings.screenshot_condition \
                and step_settings.screenshot_condition is TakeScreenshotConditionType.Inherit:
            step_settings.screenshot_condition = self.previous_settings.screenshot_condition
        self.step_settings = step_settings

    def __enter__(self):
        """Setting the driver's command_executor new settings."""
        self.driver.command_executor.settings = self.step_settings

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Reverting to the previous settings."""
        self.driver.command_executor.settings = self.previous_settings



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from time import sleep

from selenium.webdriver.remote.command import Command
from selenium.webdriver.remote.remote_connection import RemoteConnection

from src.testproject.enums import SleepTimingType


class StepHelper:

    def __init__(self, executor: RemoteConnection, w3c: bool):
        self.executor = executor
        self.w3c = w3c

    def handle_timeout(self, timeout, session_id):
        if timeout > 0:
            logging.debug(f'Setting driver implicit wait to {timeout} milliseconds.')
            if self.w3c:
                self.executor.execute(Command.SET_TIMEOUTS, {'sessionId': session_id, 'implicit': int(timeout)})
            else:
                self.executor.execute(Command.IMPLICIT_WAIT, {'sessionId': session_id, 'ms': float(timeout)})

    def handle_sleep(self, sleep_timing_type, sleep_time, command, step_executed=False):
        # Sleep Before if not Quit command
        if command is not Command.QUIT:
            if sleep_timing_type:
                sleep_timing_type_condition = SleepTimingType.After if step_executed else SleepTimingType.Before
                if sleep_timing_type is sleep_timing_type_condition:
                    logging.debug(f'Step is designed to sleep for {sleep_time} milliseconds'
                                  f' {sleep_timing_type.name} execution.')
                    sleep(sleep_time / 1000.0)



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from src.testproject.classes import ProxyDescriptor


class ActionProxy:
    """Base class that needs to be extended by custom Actions

        Attributes:
            _proxydescriptor (ProxyDescriptor): Contains the description of the custom action
    """

    def __init__(self):
        self._proxydescriptor = None

    @property
    def proxydescriptor(self) -> ProxyDescriptor:
        """Getter for the proxydescriptor"""
        return self._proxydescriptor

    @proxydescriptor.setter
    def proxydescriptor(self, value: ProxyDescriptor):
        """Setter for the proxydescriptor"""
        self._proxydescriptor = value



from .actionexecutionresponse import ActionExecutionResponse
from .elementsearchcriteria import ElementSearchCriteria
from .proxydescriptor import ProxyDescriptor
from .step_settings import StepSettings
from .driver_step_settings import DriverStepSettings

__all__ = ["ActionExecutionResponse", "ElementSearchCriteria", "ProxyDescriptor", "StepSettings",
           "DriverStepSettings"]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from selenium.webdriver.common.by import By

from src.testproject.enums import ExecutionResultType
from src.testproject.rest.messages import AddonExecutionResponse
from src.testproject.sdk.addons import ActionProxy
from src.testproject.sdk.exceptions import SdkException
from src.testproject.sdk.internal.agent import AgentClient


class AddonHelper:
    def __init__(self, agent_client: AgentClient):
        self._agent_client = agent_client

    def execute(
        self, action: ActionProxy, by: By = None, by_value: str = None
    ) -> ActionProxy:

        # Set the locator properties
        action.proxydescriptor.by = by
        action.proxydescriptor.by_value = by_value

        # Set the list of parameters for the action
        for param in action.__dict__:
            # Skip the _proxydescriptor attribute itself
            if param not in ["_proxydescriptor"]:
                action.proxydescriptor.parameters[param] = action.__dict__[param]

        response: AddonExecutionResponse = self._agent_client.execute_proxy(action)
        if response.executionresulttype != ExecutionResultType.Passed:
            raise SdkException(
                f"Error occurred during addon action execution: {response.message}"
            )

        for field in response.fields:

            # skip non-output fields
            if not field.is_output:
                continue

            # check if action has an attribute with the name of the field
            if not hasattr(action, field.name):
                logging.warning(
                    f"Action '{action.proxydescriptor.guid}' does not have a field named '{field.name}'"
                )
                continue

            # update the attribute value with the value from the response
            setattr(action, field.name, field.value)

        return action



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import inspect

from selenium.webdriver.remote.command import Command

from src.testproject.classes import StepSettings
from src.testproject.enums import TakeScreenshotConditionType
from src.testproject.helpers import ReportHelper
from src.testproject.rest.messages import DriverCommandReport, CustomTestReport
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers.redact_helper import RedactHelper


class ReportingCommandExecutor:
    """Class responsible for executing commands and reporting them

    Args:
        agent_client (AgentClient): Client used to communicate with the TestProject Agent
        command_executor: The command executor used to send WebDriver commands (Selenium or Appium)

    Attributes:
        _agent_client (AgentClient): Client used to communicate with the TestProject Agent
        _command_executor: The command executor used to send WebDriver commands (Selenium or Appium)
        _disable_reports (bool): True if all reporting is disabled, False otherwise
        _disable_auto_test_reports (bool): True if automatic reporting of tests is disabled, False otherwise
        _disable_command_reports (bool): True if driver command reporting is disabled, False otherwise
        _disable_redaction (bool): True if reporting steps should be redacted, False otherwise
        _stashed_command (DriverCommandReport): contains stashed driver command for preventing duplicates
        inside WebDriverWait
        _latest_known_test_name (str): contains latest known test name
        _excluded_test_names (list): contains a list of test names that should not be reported
    """

    def __init__(self, agent_client: AgentClient, command_executor):
        self._agent_client = agent_client
        self._command_executor = command_executor
        self._disable_reports = False
        self._disable_auto_test_reports = False
        self._disable_command_reports = False
        self._disable_redaction = False
        self._stashed_command = None
        self._latest_known_test_name = ReportHelper.infer_test_name()
        self._excluded_test_names = list()
        self._settings = StepSettings()

    @property
    def disable_reports(self) -> bool:
        """Getter for the disable_reports flag"""
        return self._disable_reports

    @disable_reports.setter
    def disable_reports(self, value: bool):
        """Setter for the disable_reports flag"""
        self._disable_reports = value

    @property
    def disable_auto_test_reports(self) -> bool:
        """Getter for the disable_auto_test_reports flag"""
        return self._disable_auto_test_reports

    @disable_auto_test_reports.setter
    def disable_auto_test_reports(self, value: bool):
        """Setter for the disable_auto_test_reports flag"""
        self._disable_auto_test_reports = value

    @property
    def disable_command_reports(self) -> bool:
        """Getter for the disable_command_reports flag"""
        return self._disable_command_reports

    @disable_command_reports.setter
    def disable_command_reports(self, value: bool):
        """Setter for the disable_command_reports flag"""
        self._disable_command_reports = value

    @property
    def disable_redaction(self) -> bool:
        """Getter for the disable_redaction flag"""
        return self._disable_redaction

    @disable_redaction.setter
    def disable_redaction(self, value: bool):
        """Setter for the disable_redaction flag"""
        self._disable_redaction = value

    @property
    def excluded_test_names(self) -> list:
        """Getter for the list of excluded test names"""
        return self._excluded_test_names

    @excluded_test_names.setter
    def excluded_test_names(self, value: list):
        """Setter for the list of excluded test names"""
        self._excluded_test_names = value

    @property
    def agent_client(self):
        """Getter for the Agent client associated with this connection"""
        return self._agent_client

    @property
    def settings(self):
        """Getter for the settings for auto step reporting defined by the user."""
        return self._settings

    @settings.setter
    def settings(self, value: StepSettings):
        self._settings = value

    def _report_command(self, command: str, params: dict, result: dict, passed: bool):
        """Reports a driver command to the TestProject platform

        Args:
            command (str): The driver command to execute
            params (dict): Named parameters to send with the command as its JSON payload
            result (dict): The response returned by the Selenium remote webdriver server
            passed (bool): True if the command execution was successful, False otherwise
        """
        if command == Command.QUIT:
            if not self.disable_auto_test_reports:
                self.report_test()
            return  # This ensures that the actual driver.quit() command is not included in the report

        if not self._disable_redaction:
            params = RedactHelper(self).redact_command(command, params)

        # If the command is executed as part of a wait loop, we don't want to report it every time
        self._is_webdriverwait = False

        # See if the command is executed inside a wait loop
        for frame in inspect.stack().__reversed__():
            if str(frame.filename).find("wait.py") > 0:
                self._is_webdriverwait = True
                break

        # Invert result is set?
        passed = not passed if self.settings.invert_result else passed

        driver_command_report = DriverCommandReport(command, params, result, passed)

        # Is screenshot needed?
        take_screenshot = False
        if self.settings.screenshot_condition is TakeScreenshotConditionType.Failure and not passed:
            take_screenshot = True
        elif self.settings.screenshot_condition is TakeScreenshotConditionType.Success and passed:
            take_screenshot = True
        elif self.settings.screenshot_condition is TakeScreenshotConditionType.Always:
            take_screenshot = True

        if take_screenshot:
            driver_command_report.screenshot = self.create_screenshot()

        if self._is_webdriverwait:
            if not self._disable_reports and not self.disable_command_reports:
                # Only stash the command for reporting later when driver command reporting is enabled
                self._stashed_command = driver_command_report
            return  # Do not report the command right away

        if not self._disable_reports and not self.disable_command_reports:
            if self._stashed_command is not None:
                # report the stashed command and clear it
                self.agent_client.report_driver_command(self._stashed_command)
                self._stashed_command = None
            # report the current command
            self.agent_client.report_driver_command(driver_command_report)

    def update_known_test_name(self):
        """Infers the current test name and if different from the latest known test name, reports a test"""
        current_test_name = ReportHelper.infer_test_name()

        # Actions inside a unittest tearDown or tearDownClass method should be reported as part of the test
        in_unittest_teardown = ReportHelper.find_unittest_teardown()

        if (
            current_test_name not in [self._latest_known_test_name, "Unnamed Test"]
            and not in_unittest_teardown
        ):
            # the name of the test method has changed and we're not inside a unittest teardown method,
            # so we need to report a test
            if not self.disable_auto_test_reports:
                self.report_test()
            # update the latest known test name for future reports
            self._latest_known_test_name = current_test_name

    def report_test(self):
        """Sends a test report to the Agent if this option is not explicitly disabled
        """

        if not self._latest_known_test_name == "Unnamed Test":

            # only report those tests that have been identified as one when their names were inferred
            if self._disable_reports:
                # test reporting has been disabled by the user
                logging.debug(f"Test [{self._latest_known_test_name}] - [Passed]")
                return

            if self._latest_known_test_name in self._excluded_test_names:
                # test has been marked as 'to be excluded, so do not report it
                logging.debug(
                    f"Test [{self._latest_known_test_name}] - Reporting skipped (marked as 'To be excluded')"
                )
                return

            custom_test_report = CustomTestReport(
                name=self._latest_known_test_name, passed=True
            )
            self.agent_client.report_test(custom_test_report)

    def create_screenshot(self) -> str:
        """Creates a screenshot (PNG) and returns it as a base64 encoded string

        Returns:
            str: The base64 encoded screenshot in PNG format (or None if screenshot taking fails)
        """
        create_screenshot_params = {
            "sessionId": self.agent_client.agent_session.session_id
        }
        create_screenshot_response = self._command_executor.execute(
            Command.SCREENSHOT, create_screenshot_params, True
        )
        try:
            return create_screenshot_response["value"]
        except KeyError as ke:
            logging.error(f"Error occurred creating a screenshot: {ke}")
            logging.error(
                f"Response from RemoteWebDriver: {create_screenshot_response}"
            )
            return None

    def clear_stash(self):
        """Reports stashed command if there is one left. Should be called when session ends to prevent
           wait-related commands from not being reported.
        """
        if not self._disable_reports and not self.disable_command_reports:
            if self._stashed_command is not None:
                # report the stashed command and clear it
                self.agent_client.report_driver_command(self._stashed_command)
                self._stashed_command = None

    def is_command_passed(self, response: dict) -> bool:
        """Determine command result based on response using state and status.

        Args:
            response (dict): The response returned by the Selenium remote webdriver server

        Returns:
            bool: True if passed, otherwise False.
        """
        # Both None and 0 response status values indicate command execution was OK
        return True if response.get("status") in [None, 0] else False
